/*****************************************************************************/
/*****************************************************************************/
/**                                LIF.C                                    **/
/**            Toolbox de fonctions d'accs aux disquettes LIF              **/
/**                                                                         **/
/**    Cration : 20/09/90          Auteur : Jacques Belin (PPC Paris #123) **/
/**    Derniere modification : 16/02/91                                     **/
/**       Correction de best_room (erreur dplacement 1er fichier)          **/
/**                                                                         **/
/*****************************************************************************/
/*****************************************************************************/
#include <bios.h>
#include <dos.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include <fcntl.h>

#if !defined(__TURBOC__)             /* MK_FP non dfini dans le MIcrosoft C */
#define MK_FP(seg, off) (unsigned long *) (((unsigned long)(seg)<<16) \
						      | (unsigned)(off))
int biosdisk(int cmd, int drive, int head, int track, int sect, int n_sect,
							      void *buffer);
#endif

/*****************************************************************************/
/*                          Codes des fichiers LIF                           */
/*****************************************************************************/
#define TEXT    1                 /* Fichier TEXT */
#define TEXT_S  -7983             /* Fichier TEXT securis */

/*************************** Fichiers HP41 ***********************************/
#define CA_41   0xe020            /* Fichier Calc */
#define XM_41   0xe030            /* Fichier X-Memory */
#define WA_41   0xe040            /* Fichier Write All */
#define KE_41   0xe050            /* Fichier Keys */
#define ST_41   0xe060            /* Fichier Status */
#define ML_41   0xe070            /* Fichier MLDL */
#define PR_41   0xe080            /* Fichier Programme */
#define BU_41   0xe0B0            /* Fichier Buffer*/
#define SDATA   -7984             /* Fichier SDATA */

/*************************** Fichiers HP71 ***********************************/
#define DATA    -7952             /* Fichier DATA */
#define DATA_S  (DATA+1)          /* Fichier DATA Scuris */
#define DATA_P  (DATA+2)          /* Fichier DATA Privatis */
#define DATA_E  (DATA+3)          /* Fichier DATA uniquement Excutable */

#define ROM71   0xe21c            /* Fichier Image Rom */
#define ROM71_S (ROM71+1)         /* Fichier Image Rom Scuris */
#define ROM71_P (ROM71+2)         /* Fichier Image Rom Privatis */
#define ROM71_E (ROM71+3)         /* Fichier Image Rom uniquement Excutable */

#define FORTH   0xe218            /* Fichier FORTH */
#define FORTH_S (FORTH+1)         /* Fichier FORTH Scuris */

#define APPT    0xe220            /* Fichier Appointements */
#define APPT_S  (APPT+1)          /* Fichier Appointements Scuris */

#define GRAPH   0xe222            /* Fichier Graphique (GRAPHLEX) */
#define GRAPH_S (GRAPH+1)         /* Fichier Graphique (GRAPHLEX) Scuris */

#define ADRS    0xe224            /* Fichier Adresses */
#define ADRS_S  (ADRS+1)          /* Fichier Adresses Scuris */

#define OBJ71   0xe22c
#define OBJ71_S (OBJ71+1)         /* Fichier DATA Scuris */

#define SYM71   0xe22e            /* Fichier Symboles (DEVROM) */
#define SYM71_S (SYM71+1)         /* Fichier Symboles (DEVROM) Scuris */

#define BIN     -7676             /* Fichier BIN */
#define BIN_S   (BIN+1)           /* Fichier BIN Scuris */
#define BIN_P   (BIN+2)           /* Fichier BIN Privatis */
#define BIN_E   (BIN+3)           /* Fichier BIN uniquement Excutable */

#define LEX     -7672             /* Fichier LEX */
#define LEX_S   (LEX+1)           /* Fichier LEX Scuris */
#define LEX_P   (LEX+2)           /* Fichier LEX Privatis */
#define LEX_E   (LEX+3)           /* Fichier LEX uniquement Excutable */

#define KEY     -7668             /* Fichier KEYS */
#define KEY_S   (KEY+1)           /* Fichier KEYS Scuris */
#define KEY_P   (KEY+2)           /* Fichier KEYS Privatis */
#define KEY_E   (KEY+3)           /* Fichier KEYS uniquement Excutable */

#define BASIC   -7660             /* Fichier BASIC */
#define BASIC_S (BASIC+1)         /* Fichier BASIC Scuris */
#define BASIC_P (BASIC+2)         /* Fichier BASIC Privatis */
#define BASIC_E (BASIC+3)         /* Fichier BASIC uniquement Excutable */


/***************************** Fichiers HP75 *********************************/
#define TEXT75  0xe052            /* Fichier TEXT */
#define APPT75  0xe053            /* Fichier Appointements */
#define IOR75   0xe058            /* Fichier Rom I/O */
#define BAS75   0xe088            /* Fichier BASIC */
#define LEX75   0xe089            /* Fichier LEX */
#define VISI75  0xe08a            /* Fichier Visicalc */
#define ROM75   0xe08b            /* Fichier Image ROM */

/*****************************************************************************/
/*  LIFDIR : structure servant  dcomposer une entre de directory          */
/*****************************************************************************/
typedef struct
          {
           char name[10];                                  /* nom du fichier */
           unsigned char type[2];                            /* code du type */
           unsigned char start[4];    /* enregistrement de dpart du fichier */
           unsigned char longueur[4];         /* longueur en enregistrements */
           unsigned char date[6];                       /* groupe date/heure */
           unsigned char flag[2];             /* flag interne (valeur 8001h) */
           unsigned char applic[4];  /* valeurs dpendant du type de fichier */
          } LIFDIR;

/*****************************************************************************/
/*  GDH : structure Groupe Date/Heure                                        */
/*****************************************************************************/
typedef struct
         {
          int day;                                                   /* jour */
          int month;                                                 /* mois */
          int year;                                                 /* anne */
          int hour;                                                 /* heure */
          int min;                                                /* minutes */
          int sec;                                               /* secondes */
         } GDH;

/*****************************************************************************/
/*                       Variables Globales communes                         */
/*****************************************************************************/
char buff_sect[256];                   /* buffer contenant un enregistrement */
int curr_sect;                                  /* numro de secteur courant */
int curr_dir;                    /* numro de l'entre de directory courante */
int curr_char;                    /* numro de l'octet point dans buff_sect */
unsigned long far *intptr;      /* pointeurs servant  accder  la table .. */
char far *descptr;              /* .. de description du lecteur de disquette */
char lsect;          /* sauvegarde valeur initiale de la longueur de secteur */



/*****************************************************************************/
/*   INIT_DSK : Initialisation du lecteur de disquette                       */
/*             A insrer au dbut du programme, pour permettre la            */
/*             reconnaissance du formatage des disquette LIF                 */
/*   entres : aucune                                                        */
/*   utilise : intptr, descptr, lsect (variables globales)                   */
/*   valeur retourne : aucune                                               */
/*****************************************************************************/
void init_dsk(void)
   {
    intptr=MK_FP(0x0000, 0x0078);    /* fabrication du pointeur sur la table */
    descptr=(void far *) *intptr;                        /* accs  la table */
    lsect=descptr[3];        /* sauvegarde de l'ancienne longueur de secteur */
    descptr[3]=1;                            /* nouvelle longueur de secteur */
   }


/*****************************************************************************/
/*   FIN     : fin du programme                                              */
/*             A utiliser en fin de programme,  la place de "exit", pour    */
/*             permettre, ensuite, la reconnaissance des disquettes MS/DOS   */
/*   entres : code de sortie                                                */
/*   utilise : intptr, descptr, lsect (variables globales)                   */
/*   valeur retourne : aucune                                               */
/*****************************************************************************/
void fin(int code)
   {
    descptr[3]=lsect;        /* remise en place ancienne longueur de secteur */
    exit(code);                                       /* sortie du programme */
   }


/*****************************************************************************/
/*   READ_DSK : lecture d'un enregistrement de disquette LIF                 */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             secteur = numro de l'enregistrement (0 pour le premier)      */
/*   retours : buffer = enregistrement venant d'tre lu                      */
/*   valeur returne : retour int 13h : si different de 0 -> erreur          */
/*****************************************************************************/
int read_dsk(int drive, int secteur, char *buffer)
   {
    int retour=-1;
    int i;
    union REGS inregs, outregs;
    int head, track, sect;

           /* dcomposition du secteur logique en ses constituants physiques */
    head= (secteur>>4) & 0x0001;                        /* (secteur/16) % 2) */
    track=secteur>>5;                                          /* secteur/32 */
    sect=(secteur & 0x000f)+1;                               /* secteur % 16 */

    for(i=1; (i<=3) && (retour != 0); i++)          /* on s'accorde 3 essais */
      {
       retour=biosdisk(2, drive, head, track, sect, 1, buffer);   /* lecture */
       if(retour != 0)                                          /* si erreur */
         {
          inregs.h.ah=0x00;
	  inregs.h.dl=(char) drive;
          int86(0x13, &inregs, &outregs);     /* rinitialisation du lecteur */
         }
      }
    return(retour);
   }


/*****************************************************************************/
/*   WRT_DSK : Ecriture d'un enregistrement de disquette LIF                 */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             secteur = numro de l'enregistrement (0 pour le premier)      */
/*   retours : buffer = enregistrement  crire                              */
/*   valeur returne : retour int 13h : si different de 0 -> erreur          */
/*****************************************************************************/
int wrt_dsk(int drive, int secteur, char *buffer)
   {
    int retour=-1;
    int i;
    union REGS inregs, outregs;
    int head, track, sect;

    head= (secteur>>4) & 0x0001;                        /* (secteur/16) % 2) */
    track=secteur>>5;                                          /* secteur/32 */
    sect=(secteur & 0x000f)+1;                               /* secteur % 16 */
    for(i=1; (i<=4) && (retour != 0); i++)
      {
       retour=biosdisk(3, drive, head, track, sect, 1, buffer);
       if(retour != 0)
         {
          inregs.h.ah=0x00;
	  inregs.h.dl=(char) drive;
          int86(0x13, &inregs, &outregs);
         }
      }
    if(retour == 3)
      {
       fprintf(stderr, "Error : disk protected\n");
       fin(-1);
      }
    return(retour);
   }


/*****************************************************************************/
/*   READ_CHAR : Lecture squentielle de caracteres sur disque LIF           */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             curr_sect, curr_dir = variables globales,  initialiser       */
/*               avant la premire utilisation : curr_sect=255;              */
/*               curr_dir=numero du premier secteur  lire                   */
/*   utilise : curr_sect, curr_dir, buff_sect (variables globales)           */
/*   valeur retourne : caractere suivant sur le disque                      */
/*****************************************************************************/
unsigned char read_char(int drive)
   {
    if(curr_char == 255)    /* si dernier caractere lu dans l'enregistrement */
      {
       read_dsk(drive, curr_sect, (char *) buff_sect);     /* lecture enreg. */
       curr_sect++;                            /* incrmentation num secteur */
       curr_char=-1;                              /* initialisation pointeur */
      }
    return(buff_sect[++curr_char]);             /* lecture caractre courant */
   }


/*****************************************************************************/
/*   WRITE_CHAR : Ecriture squentielle de caracteres sur disque LIF         */
/*   Note : cette fonction utilise la variable globale "buff_sect" comme     */
/*          buffer. Il faut donc prendre des prcautions :                   */
/*          1 : Vous devez crire vous-mme le contenu final du buffer       */
/*            en fin de la boucle de traitement (avec wrt_dsk).              */
/*          2 : n'utilisez pas de fonction modifiant "buff_sect" avant       */
/*            l'criture finale.                                             */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             c = caractere  ecrire sur disque                             */
/*   utilise : curr_sect=numro du premier secteur  crire,                 */
/*             curr_char  initialiser  0 au dbut du traitement            */
/*             buff_sect (variables globales)                                */
/*   valeur retourne : 0, en cas d'erreur code de retour de int 13h         */
/*****************************************************************************/
int write_char(int drive, unsigned char c)
   {
    int retour=0;

    buff_sect[curr_char++]=c;        /* ecriture du caractre dans le buffer */
    if(curr_char == 256)                          /* si fin d'enregistrement */
      {
       retour=wrt_dsk(drive, curr_sect, (char *) buff_sect); /* ecrit. buff. */
       curr_sect++;                         /* incrmentation numro secteur */
       curr_char=0;                   /* rinitialisation pointeur caractre */
      }
    return(retour);
   }


/*****************************************************************************/
/*   READ_DIR : Lecture squentielle du rpertoire                           */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             curr_sect, curr_dir = variables globales,  initialiser       */
/*               avant la premire utilisation : curr_sect=1; curr_dir=7;    */
/*   retours : buffer = Structure LIFDIR contenant les donnes de l'entre   */
/*               du rpertoire venant d'tre lu                              */
/*   utilise : curr_sect, curr_dir, buff_sect (variables globales)           */
/*   valeur retourne : 1, ou 0 aprs la lecture de la fin du rpertoire     */
/*****************************************************************************/
int read_dir(int drive, LIFDIR *buffer)
   {
    int retour=1;

    if(++curr_dir == 8)                           /* si fin d'enregistrement */
      {
       read_dsk(drive, ++curr_sect, buff_sect);    /* lecture enreg. suivant */
       curr_dir=0;         /* curr_dir pointe sur la 1ere entre de l'enreg. */
      }
    memcpy(buffer, buff_sect+(curr_dir<<5), 32); /* copie de l'entre ds buff*/

    if((buffer->type[0]==255) && (buffer->type[1]==255)) /* si fin de direct.*/
       retour=0;
    return(retour);
   }


/*****************************************************************************/
/*   SEEK_DIR : Acces direct  une entre du rpertoire                      */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             pos = numro de l'entre  lire (0 pour la premire)          */
/*   retours : dir = Structure LIFDIR contenant les donnes de l'entre      */
/*               du rpertoire venant d'tre lu                              */
/*   utilise : buff_sect (variable globale)                                  */
/*   valeur retourne : 0, ou -1 en cas d'erreur                             */
/*****************************************************************************/
int seek_dir(int drive, int pos, LIFDIR *dir)
   {
    unsigned int pos_record;         /* enregistrement correspondant a 'pos' */
    unsigned int pos_dir;         /* position de 'pos' dans l'enregistrement */

    pos_record = (pos/8)+2;           /* dcomposition du numro de l'entre */
    pos_dir = pos % 8;
                               /* lecture du premier secteur de la disquette */
    if(read_dsk(drive, 0, buff_sect))
       return(-1);
                            /* si dpassement de la longueur de directory... */
    if((((int)buff_sect[18]<<16)+buff_sect[19]) <= (pos_record-2))
      {
       return(-1);                                             /* ... erreur */
      }
    read_dsk(drive, pos_record, buff_sect); /* lecture secteur correspondant */
    memcpy(dir, buff_sect+(32*pos_dir), 32);     /* copie buff_sect dans dir */
    return(0);
   }


/*****************************************************************************/
/*   SCH_DIR : recherche de l'entre du rpertoire, suivant nom du fichier   */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             name = nom du fichier  trouver, sur 10 caractres (avec      */
/*               des caractres ' ' en fin de chaine)                        */
/*   retours : pos_dir = numro de l'entre du rpertoire trouve (0 pour    */
/*               la premire)                                                */
/*             dir = Structure LIFDIR contenant les donnes de l'entre      */
/*               du rpertoire venant d'tre lu                              */
/*   utilise : curr_sect, curr_dir, buff_sect (variables globales)           */
/*   valeur retourne : si = -1 : fichier non trouv                         */
/*                      si = -2 : fichier trouv mais effac                 */
/*                      si >  0 : longueur fichier en enregistrements        */
/*****************************************************************************/
int sch_dir(int drive, char *name, int *pos_dir, LIFDIR *dir)
   {
    int  retour=-1;                                           /* code retour */
    int  etat;                          /* etat du fichier (si = 0 : effac) */
    int  found=0, not_end;   /* flag d'arret de recherche, indic. fin de dir */

    curr_sect=1;    /* initialisation avant la premire lecture du directory */
    curr_dir=7;
    do
      {
       not_end=read_dir(drive, dir);    /* lecture squentielle du directory */
       if(not_end)                                       /* si entre valide */
         {
          etat=((int)dir->type[0]<<8)+dir->type[1];       /* lecture du type */
          if(strncmp(dir->name, name, 10) == 0)         /* si fichier trouv */
            {
             if(etat)                      /* si le fichier n'est pas effac */
               {
                found=1;                         /* on arretera la recherche */
                                                    /* calcul de la longueur */
                retour=((int)dir->longueur[2]<<8)+dir->longueur[3];
               }
             else                                       /* si fichier effac */
               {
                retour=-2;                  /* on positione le retour  '-2' */
               }
             *pos_dir=((curr_sect-2)*8)+curr_dir;   /* calcul pos. directory */
            }
         }
      }
    while(not_end && !found); /* boucle tant que le fichier n'est pas trouve */

    return(retour);
   }


/*****************************************************************************/
/*   WRT_DIR : Ecriture d'une entre de rpertoire                           */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             pos = numro de l'entre  crire (0 pour la premire)        */
/*             dir = Structure LIFDIR contenant les donnes de l'entre      */
/*               du rpertoire  crire                                      */
/*   utilise : buff_sect (variable globale)                                  */
/*   valeur retourne : 0, ou -1 en cas d'erreur                             */
/* NOTE : cette fonction normalise le premier enregistrement du directory    */
/*****************************************************************************/
int wrt_dir(int drive, int pos, LIFDIR *dir)
   {
    unsigned int pos_record;
    unsigned int pos_dir;

    pos_record = (pos/8)+2;
    pos_dir = pos % 8;

    if(read_dsk(drive, 0, buff_sect))
       return(-1);
    if((((int)buff_sect[18]<<8)+buff_sect[19]) <= (pos_record-2))
      {
       return(-1);
      }
    read_dsk(drive, pos_record, buff_sect);     /* lecture secteur directory */
    memcpy(buff_sect+(32*pos_dir), dir, 32);  /* copie entre dans buff_sect */
    if(pos < 7)           /* normalisation premier enregistrement rpertoire */
       if(buff_sect[32*(pos_dir+1)] == 0)
          memset(buff_sect+(32*(pos_dir+1)), 0xFF, (7-pos_dir)*32);
    wrt_dsk(drive, pos_record, buff_sect);             /* recriture secteur */
    return(0);
   }


/*****************************************************************************/
/*   CLR_FILE: effacement d'un fichier (en annulant son type)                */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             pos = numro de l'entre  effacer (0 pour la premire)       */
/*   utilise : buff_sect (variable globale)                                  */
/*   valeur retourne : 0, ou -1 en cas d'erreur                             */
/*****************************************************************************/
int clr_file(int drive, int pos)
   {
    unsigned int pos_record;
    unsigned int pos_dir;

    pos_record = (pos/8)+2;
    pos_dir = pos % 8;
    if(read_dsk(drive, 0, buff_sect))
       return(-1);
    if((((int)buff_sect[18]<<8)+buff_sect[19]) <= (pos_record-2))
      {
       return(-1);
      }
    read_dsk(drive, pos_record, buff_sect);               /* lecture secteur */
    memset(buff_sect+(32*pos_dir)+10, 0, 2);        /* remise  zero du type */
    wrt_dsk(drive, pos_record, buff_sect);             /* recriture secteur */
    return(0);
   }


/*****************************************************************************/
/*  BUILD_DIR : construction d'une entre de directory                       */
/*  Entres = name = nom du fichier (sur 10 caracteres)                      */
/*            private = code de protection du fichier. 1 ou 0                */
/*            secure = code de securisation du fichier. 1 ou 0               */
/*            type = code numrique du type de fichier                       */
/*            start = numro du 1er enregistrement du fichier                */
/*            long_file = longueur du fichier en enregistrement              */
/*            gdh = structure GDH contenant la date et l'heure du fichier    */
/*            imp1, imp2 = valeurs dpendant du type de fichier (par         */
/*              exemple imp1=longueur fich en nibbles pour LEX, BASIC...)    */
/*  Sortie = structure LIFDIR contenant les donnes formates (avec la date  */
/*             courante)                                                     */
/*  Valeur retourne : aucune                                                */
/*****************************************************************************/
void build_dir(LIFDIR *dir, char *name, int private, int secure, int type,
                            int start, int long_file, GDH *gdh, long imp1,
                            long imp2)
    {
     memset(dir, 0, 32);          /* initialisation de la zone mmoire  '0' */
     memcpy(dir->name, name, 10);                 /* copie du nom du fichier */

                                      /* enregistrement de dpart du fichier */
     dir->start[2]=(unsigned char) (start >> 8);   
     dir->start[3]=(unsigned char) (start & 0xff);

                                              /* longueur en enregistrements */
     dir->longueur[2]=(unsigned char) (long_file >> 8);   
     dir->longueur[3]=(unsigned char) (long_file & 0xff);

     gdh->year %= 100;                             /* ecriture date et heure */
     dir->date[0]=(char) ((gdh->year/10)*16)+(gdh->year % 10);
     dir->date[1]=(char) ((gdh->month/10)*16)+(gdh->month % 10);
     dir->date[2]=(char) ((gdh->day/10)*16)+(gdh->day % 10);
     dir->date[3]=(char) ((gdh->hour/10)*16)+(gdh->hour % 10);
     dir->date[4]=(char) ((gdh->min/10)*16)+(gdh->min % 10);
     dir->date[5]=(char) ((gdh->sec/10)*16)+(gdh->sec % 10);

     dir->flag[0]=0x80;                                /* flag disquette LIF */
     dir->flag[1]=0x01;

                                /* constitution du types et des codes 'applic'
                            suivant les types de fichiers et les protections */
     switch (type)
       {
        case TEXT  : if(secure)
                        type=TEXT_S;
                     break;


        case FORTH :
        case ROM71 :
        case APPT  :
        case GRAPH :
        case ADRS  :
        case OBJ71 :
        case SYM71 :

        case BIN   :
        case LEX   :
        case KEY   :
        case BASIC : type += ((private<<1)+secure);
                     dir->applic[0]=(unsigned char) (imp1 & 0xff);
                     dir->applic[1]=(unsigned char) ((imp1>>8) & 0xff);
                     dir->applic[2]=(unsigned char) (imp1>>16);
                     break;

        case DATA  : type += ((private<<1)+secure);
                     dir->applic[0]=(unsigned char) (imp1 & 0xff);
                     dir->applic[1]=(unsigned char) (imp1>>8);
                     dir->applic[2]=(unsigned char) (imp2 & 0xff);
                     dir->applic[3]=(unsigned char) (imp2>>8);
                     break;


        case CA_41 :
        case XM_41 :
        case ML_41 :
        case BU_41 :

        case PR_41 :
        case WA_41 :
        case KE_41 :
        case ST_41 :
        case SDATA : dir->applic[0]=(unsigned char) (imp1 >> 8);
                     dir->applic[1]=(unsigned char) (imp1 & 0xff);
                     dir->applic[2]=(unsigned char) ((secure<<4) + private);
                     break;

       }
     dir->type[0]=(unsigned char) (type>>8);              /* type du fichier */
     dir->type[1]=(unsigned char) (type & 0xff);
    }


/*****************************************************************************/
/*  BREAK_DIR : dcomposition d'une entre de directory                      */
/*  Entre = structure LIFDIR contenant les donnes formates (avec la date  */
/*             courante)                                                     */
/*  Sorties = name = nom du fichier (sur 10 caracteres)                      */
/*            private = code de protection du fichier. 1 ou 0                */
/*            secure = code de securisation du fichier. 1 ou 0               */
/*            type = code numrique du type de fichier                       */
/*            start = numro du 1er enregistrement du fichier                */
/*            long_file = longueur du fichier en enregistrement              */
/*            gdh = structure GDH contenant la date et l'heure du fichier    */
/*            imp1, imp2 = valeurs dpendant du type de fichier (par         */
/*              exemple imp1=longueur fich en nibbles pour LEX, BASIC...)    */
/*  Valeur retourne : aucune                                                */
/*****************************************************************************/
void break_dir(char *name, int *private, int *secure, int *type, int *start,
            int *long_file, GDH *gdh, long *imp1, long *imp2, LIFDIR *dir)
    {
     memcpy(name, dir->name, 10);                            /* copie du nom */
     name[10]='\0';

     *type=((int)dir->type[0]<<8)+dir->type[1];            /* calcul du type */

     *secure=0;
     switch(*type)            /* determination du flag 'secure' suivant type */
       {
        case TEXT_S  : *secure = 1;
                       *type = TEXT;
                       break;


        case CA_41   :
        case XM_41   :
        case ML_41   :
        case BU_41   :

        case PR_41   :
        case WA_41   :
        case KE_41   :
        case ST_41   :
        case SDATA   : if(dir->applic[2] & 8)
                          *secure = 1;
                       break;


        case BIN_E   :
        case LEX_E   :
        case BASIC_E :
        case ROM71_E :

        case ROM71_S :
        case FORTH_S :
        case APPT_S  :
        case GRAPH_S :
        case ADRS_S  :
        case OBJ71_S :
        case SYM71_S :
        case BIN_S   :
        case LEX_S   :
        case DATA_S  :
        case KEY_S   :
        case BASIC_S : *secure = 1;
                       *type &= 0xfffe;
                       break;

        default      : break;
       }

     *private=0;
     switch(*type)          /* dtermination du flag 'private ' suivant type */
       {
        case CA_41   :
        case XM_41   :
        case ML_41   :
        case BU_41   :

        case PR_41   :
        case WA_41   :
        case KE_41   :
        case ST_41   :
        case SDATA   : if(dir->applic[2] & 1)
                          *private = 1;
                       break;

        case ROM71_P :
        case BIN_P   :
        case LEX_P   :
        case BASIC_P : *private = 1;
                       *type &= 0xfffd;
                       break;

        default      : break;
       }

     *start=((int)dir->start[2] << 8) + dir->start[3];      /* enreg. dpart */

     *long_file=((int)dir->longueur[2] << 8) + dir->longueur[3]; /* longueur */

     gdh->year=((dir->date[0]/16)*10) + (dir->date[0] % 16);  /*date et heure*/
     gdh->month=((dir->date[1]/16)*10) + (dir->date[1] % 16);
     gdh->day=((dir->date[2]/16)*10) + (dir->date[2] % 16);
     gdh->hour=((dir->date[3]/16)*10) + (dir->date[3] % 16);
     gdh->min=((dir->date[4]/16)*10) + (dir->date[4] % 16);
     gdh->sec=((dir->date[5]/16)*10) + (dir->date[5] % 16);


     *imp1 = 0L;
     *imp2 = 0L;
     switch(*type)      /* calculs des valeurs dpendant du type de fichiers */
       {
        case ROM71 :
        case FORTH :
        case APPT  :
        case GRAPH :
        case ADRS  :
        case OBJ71 :
        case SYM71 :

        case BIN   :
        case LEX   :
        case KEY   :
        case BASIC : *imp1 = ((long)dir->applic[2]<<16) +
                             ((long)dir->applic[1]<<8) + dir->applic[0];
                     break;

        case DATA  : *imp1 = ((long)dir->applic[1]<<8)+dir->applic[0];
                     *imp2 = ((long)dir->applic[3]<<8)+dir->applic[2];
                     break;

        case CA_41 :
        case XM_41 :
        case ML_41 :
        case BU_41 :

        case PR_41 :
        case WA_41 :
        case KE_41 :
        case ST_41 :
        case SDATA : *imp1=((long)dir->applic[0]<<8)+dir->applic[1];
                     break;

       }
    }


/*****************************************************************************/
/*   BEST_ROOM: recherche de places pour crire un fichier                   */
/*   entres : drive = numro du lecteur : '0' pour le lecteur A             */
/*             name = nom du fichier  crire,  sur 10 caractres (avec      */
/*               des caractres ' ' en fin de chaine)                        */
/*             long_file = longueur du fichier, en enregistrements           */
/*   retours : pos_dir = numro de l'entre du rpertoire trouve (0 pour    */
/*               la premire)                                                */
/*             cl_dir = ventuellement (si different de -1), numro de       */
/*               l'ancienne entre en cas de dplacement de fichier.         */
/*               Il faudra en effacer le type                                */
/*   utilise : curr_sect, curr_dir, buff_sect (variables globales)           */
/*   valeur retourne : si = -1 : plus de place pour le fichier              */
/*                      si = -2 : plus de place dans le rpertoire           */
/*                      si = -3 : erreur acces disque                        */
/*                      si >  0 : numro du 1er enreg. pour le fichier       */
/*****************************************************************************/
int best_room(int drive, char *name, int long_file, int *pos_dir, int *cl_dir)
   {
    int    longd;                   /* nombre maximum d'entres du directory */
    int    first_file;                        /* position du premier fichier */
    LIFDIR dir;                     /* structure contenant l'entre courante */
    GDH    gdh;                               /* structure groupe-date-heure */
    char   buffer[256];                        /* contenu du secteur courant */
    int    i=0;                              /* numro de l'entree courante */
    int    not_end=1;                      /* indicateur de fin de directory */
    int    last_file=-1;     /* numro de l'entre du dernier fichier valide */
    int    hole=-2;     /* numro de secteur du plus petit 'trou' convenable */
    int    s_hole=9999;            /* taille du plus petit 'trou' convenable */
    int    file_ok=0; /* flag indiquant 'fichier trouv et taille identique' */
    int    file_dim=0;  /* flag indiquant que le fichier a diminu de taille */
    int    file_size;                 /* taille du fichier en enregistrement */
    int    file_start;                  /* premier enregistrement du fichier */
    int    new_dir=0;                             /* nouveau numro d'entre */
    int    new_sect=0;                          /* nouveau secteur de dpart */
    int    retour=-1;                    /* valeur retourne par la fonction */

    *cl_dir=-1;
    if(read_dsk(drive, 0, buffer))    /* lecture de l'entte de la disquette */
       return(-3);                                 /* fin fonction si erreur */

                                     /* calcul longueur maximum du directory */
    longd = (((int) buffer[18]<<8)+(int) buffer[19])*8;
    first_file = ((int) buffer[18]<<8)+(int) buffer[19]+2;

    curr_sect=1;
    curr_dir=7;
    while(read_dir(drive, &dir) && not_end) /* lecture squentielle directory*/
      {
       if((dir.type[0]==255) && (dir.type[1]==255))      /* si fin directory */
         {
          not_end=0;                               /* on terminera la boucle */
         }
       else                                                      /* sinon... */
         {
                        /* calcul taile (en enreg.) du fichier correspondant */
          file_size=((int)dir.longueur[2]<<8)+dir.longueur[3];
          if((dir.type[0]==0) && (dir.type[1]==0))      /* si fichier effac */
            {
             if(file_size >= long_file)  /*si taille ancien >= taille nouveau*/
               {
                if(file_size < s_hole)  /* si taile ancien < taile+grand trou*/
                  {
                   s_hole=file_size;        /* ce trou est plus interressant */
                   hole=i;
                  }
               }
            }
          else                                      /* si fichier non effac */
            {
             last_file=i;
             if(strncmp(dir.name, name, 10) == 0)       /* si fichier trouv */
               {
                if(file_size < long_file)    /* si fichier devenu trop grand */
                  {
                   *cl_dir=i;              /* il faudra effacer cette entre */
                  }
                if(file_size == long_file)                  /* si fichier ok */
                  {
                   file_ok=1;                            /* fin de la boucle */
                   *pos_dir=i;
                   not_end=0;
                  }
                if(file_size > long_file)            /* si fichier  retreci */
                  {
                   *cl_dir=i;   /* il faudra (probablement) effacer l'entre */
                   file_dim=1;                            /* on arme le flag */
                  }
               }
            }
          i++;                          /* incrmentation du numro d'entre */
         }
      }

    if(file_ok)                                      /* si pas de changement */
      {
       seek_dir(drive, *pos_dir, &dir);  /*acquisition donnes entre trouve*/
       retour=((int)dir.start[2]<<8)+dir.start[3];   /* calcul pos 1er enreg */
      }
    else                                      /* recherche d'une autre place */
      {
       if((s_hole==long_file) && (hole<last_file)) /*si trou correspond exact*/
         {
          *pos_dir=hole;               /* on placera le fichier dans ce trou */
          seek_dir(drive, hole, &dir);
          retour=((int)dir.start[2]<<8)+dir.start[3];
         }
       else                                                         /* sinon */
         {
          i=last_file;              /* recherche derniere entre non efface */
          if(i < 0)                                     /* si disquette vide */
                                      /* simulation d'une entre numro '-1' */
	     build_dir(&dir, "", 0, 0, 0x0000, first_file-1, 1, &gdh, 0L, 0L);
          else
             seek_dir(drive, i, &dir); /*rcupration donnes derniere entre*/

          file_start=((int)dir.start[2]<<8)+dir.start[3]; /* enr. depart fich*/

                                         /* si le dernier fichier est le notre,
                                  et qu'il reste suffisemment de place apres */
          if((strncmp(dir.name, name, 10) == 0) &&
             ((file_start+long_file) <= 2464))
            {
             *cl_dir=-1;  /* on annule l'effacement de l'entre, activ+haut */
             *pos_dir=i;             /* positionnement des valeurs en sortie */
             retour=file_start;
            }
          else
            {
             new_dir=i+1;                       /* recherche entre suivante */
             if(i>=0)
                new_sect=file_start+((int)dir.longueur[2]<<8)+dir.longueur[3];
             else
                new_sect=first_file;
             if((new_dir < longd) && ((new_sect+long_file) <= 2464))
               {                          /* si place, enregistrement en fin */
                if(i<0)
                   *pos_dir=0;                   /* c'est le premier fichier */
                else
                   *pos_dir=new_dir;            /* c'est une nouvelle entre */
                retour=new_sect;
               }
             else    /* si plus de place en fin, on cherche les grands trous */
               {
                if(file_dim)                      /* si le fichier a rtrci */
                  {           /* on le rcrira tout de mme  la mme place */
                   seek_dir(drive, *cl_dir, &dir);
                   *pos_dir=*cl_dir;
                   *cl_dir=-1;
                   retour=((int)dir.start[2]<<8)+dir.start[3];
                  }
                else                                /* la derniere chance... */
                  {
                   if(hole)  /* si un trou (mme beaucoup plus grand) existe */
                     {
                      seek_dir(drive, hole, &dir); /* on y met notre fichier */
                      *pos_dir=hole;
                      retour=((int)dir.start[2]<<8)+dir.start[3];
                     }
                  }
               }
            }
         }
      }
    return(retour);
   }
