/*****************************************************************************/
/*****************************************************************************/
/**                            LIFREAD.C                                    **/
/**            Dump d'enregistrements de disquette LIF                      **/
/**                                                                         **/
/**    Cration : 04/10/90               Auteur : JB                        **/
/**    Derniere modification : 06/02/91                                     **/
/**       Modification de l'analyse des paramtres                          **/
/*****************************************************************************/
/*****************************************************************************/
#include <ctype.h>
#include <conio.h>
#include "getopt.c"
#include "lif.c"

#if defined(__TURBOC__)
#include "lif.tc"
#else
#include "lif.msc"
#endif

#define COL 16              /* 16 lignes * 16 colonnes = 256 octest affichs */
#define LIG 16

int conv_parm(char *parm);

void main(int argc, char *argv[])
   {
    int lif_drive = 0;                   /* numro du lecteur (0 par dfaut) */
    int err_entry = 0;                                /* indicateur d'erreur */
    int start, end;                /* premier et dernier enregistrements lus */
    int retour;                  /* code de retour de la fonction de lecture */
    int i, j, k;
    char c;


    SW='-';
    while((c=(char) getopt(argc, argv, "d:")) != EOF)
      {
       switch(c)
         {
          case 'd' : lif_drive = *optarg-'0';        /* lecteur de disquette */
                     if((lif_drive != 0) && (lif_drive != 1))
                        err_entry++; 
                     break;

          case '?' : err_entry++;                               /* si erreur */
                     break;
         }
      }

    if(!err_entry)
      {
       switch(argc-optind)            /* analyse des paramtres start et end */
         {                                     /* il y a 1 paramtre : start */
          case 0  : start = 0;
                    end = 0;
                    break;

          case 1  : start=conv_parm(argv[optind]);   /* conversion paramtre */
                    if(start == -1)        /* si paramtre invalide : erreur */
                       err_entry = 1;
                    end=start;    /* on ne lira que l'enregistrement 'start' */
                    break;

          case 2  : start=conv_parm(argv[optind]);
                    if(start == -1)
                       err_entry = 1;
                    end=conv_parm(argv[optind+1]);
                    if(end == -1)
                       err_entry = 1;
                    break;
                                 /* nombre de paramtres diffrents : erreur */
          default : err_entry = 1;
                    break;
         }

       if(start > end)                 /* controle de l'ordre des paramtres */
          err_entry = 1;
      }

    if(err_entry)                                               /* si erreur */
      {
       fprintf(stderr, "usage : lifread [-d drive_number] [start [end]]\n");
       exit(-1);
      }

    init_dsk();                    /* initialisation du lecteur de disquette */

               /* lecture de 'start'  'end' ou jusqu' l'appui d'une touche */
    for(k=start; (k<=end) && (!kbhit()); k++)
      {
       fprintf(stdout, "\nsector %d\n", k);   /* affichage numro de secteur */
       retour=read_dsk(lif_drive, k, (char *) buff_sect); /* lecture secteur */
       if(retour)                                    /* si erreur de lecture */
         {
          switch(retour)        /* affichage des diffrentes causes d'erreur */
            {
             case 0x02 :  fprintf(stdout, "Adress mark not found\n");
                          break;

             case 0x04 :  fprintf(stdout, "Sector not found\n");
                          break;

             case 0x20 :  fprintf(stdout, "Disk controler Error\n");
                          break;

             case 0x40 :  fprintf(stdout, "Track not found\n");
                          break;

             case 0x80 :  fprintf(stdout,"No response from the disk reader\n");
                          break;

             default   :  fprintf(stdout, "Read error : return code = %2Xh\n",
                                                                      retour);
                          break;
            }
         }
       else                                          /* si lecture effectue */
         {
          for(i=0; i<=LIG-1; i++)                /* traitement des 16 lignes */
            {
                                            /* affichage des octets en Hexa */
             for(j=COL*i; j<=(COL*i+(COL-1)); j++) /*traitement de 16 octets */
               {
                if((j % 4) == 0)   /* un espace suplmentaire aprs 4 octets */
                   fprintf(stdout, " ");
                                      /* affichage de l'octet en hexadcimal */
                fprintf(stdout, "%02X ", buff_sect[j] & 0x00ff);
               }
             fprintf(stdout, "    "); /* quelques espaces pour faire beau... */
                                          /* affichage des octets 'en clair' */
             for(j=COL*i; j<=(COL*i+(COL-1)); j++) /*traitement de 16 octets */
               {
                if((buff_sect[j] < 32) ||      /* si caractere non 'visible' */
                  ((unsigned char) buff_sect[j] >= 128))
                   fprintf(stdout, ".");             /* affichage d'un point */
                else                                 /* si caractre visible */
                   fprintf(stdout, "%c", buff_sect[j]); /*affichage caractre*/
               }
             fprintf(stdout, "\n");                       /* fin de la ligne */
            }
         }
      }
    fin(0);               /* fin du programme et rinitialisation du lecteur */
   }



/*****************************************************************************/
/*   CONV_PARM : conversion et controle de paramtre                         */
/*   entres : paramtre sous forme de chaine alpha-numrique                */
/*   valeur retourne : param sous forme dcimale, ou -1 si erreur           */
/*****************************************************************************/
int conv_parm(char *parm)
   {
    int res=0;                                                   /* rsultat */
    int i=0;

    if(strlen(parm) > 4)                 /* si plus de 4 caractres : erreur */
      {
       res=-1;
      }
    else                                                            /* sinon */
      {
       while((parm[i] != '\0') && (res != -1)) /*analyse de chaque caractre */
         {
          if(isdigit(parm[i]))                     /* si caractre numrique */
             res = res*10 + parm[i]-'0';             /* calcul nouveau 'res' */
          else                                               /* sinon erreur */
             res=-1;
          i++;
         }
       if(res > 2463)                     /* si numro de secteur trop grand */
          res=-1;
      }
    return(res);                                   /* retour valeur calcule */
   }
